"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import cached_property, partial
from typing import Any
from pydantic import Field, field_validator
import numpy as np


from ..eka import Circuit
from ..eka.utilities import BoolOp

from .op_signature import (
    BOOL_LOGIC_OP_SIGNATURE,
    CONTROL_FLOW_OP_SIGNATURE,
    USUAL_QUANTUM_GATES,
    NONCLIFFORD_GATES_SIGNATURE,
    UTILS_SIGNATURE,
    OpType,
    OpSignature,
)
from .converter import Converter, OpAndTargetToInstrCallable

# Alias for the output of a PennyLane/catalyst run.
PennyLaneResult = dict[str, Any]


class EkaToPennylaneConverter(Converter[str, PennyLaneResult]):
    """Convert Eka InterpretationStep to PennyLane.

    Here's a simple example of how to use this method to execute Eka experiment with
    PennyLane:

    .. code-block:: python

        from loom.executor import EkaToPennylaneConverter

        pl_converter = EkaToPennylaneConverter(is_catalyst=False)
        pl_program, q_register, c_register = pl_converter.convert(interpreted_eka)
        n_qubits = len(q_register)

        # Indent the program body so it fits inside a function
        indented_program = "\\n    ".join(pl_program.splitlines())

        # Construct the Python program string
        s_prog = f\"\"\"
        import pennylane as qml

        def circuit():
            {indented_program}
            return {{k:qml.sample(measurements[k]) for k in measurements.keys()}}

        # Use a PennyLane device
        dev = qml.device("lightning.qubit", wires=n_qubits, shots=5)
        circ = qml.QNode(circuit, dev)

        results = circ()
        \"\"\"
        local_ns = {}
        exec(s_prog, globals(), local_ns)

        results = local_ns["results"]

        parsed_outcome = pl_converter.parse_target_run_outcome(results)

    Parameters
    ----------
    is_catalyst : bool
        Whether the PennyLane program is intended to run on Catalyst.
    import_prefix : str
        The import alias for PennyLane defaults to "qml.".
    """

    SUPPORTED_OPERATIONS: frozenset[OpSignature] = USUAL_QUANTUM_GATES

    ALLOW_ERROR_MODELS: bool = Field(default=False, frozen=True, init=False)

    # Specify whether the circuit is meant to run on catalyst or not.
    is_catalyst: bool = Field(default=False, frozen=True, init=True)

    import_prefix: str = Field(
        default="qml.",
        frozen=True,
        init=True,
        description="The import alias for PennyLane.",
    )

    @field_validator("import_prefix")
    # pylint: disable=no-self-argument
    def validate_import_prefix(cls, v: str) -> str:
        """Ensure the import prefix ends with a dot if not empty."""
        if v and not v.endswith("."):
            v += "."
        return v

    @cached_property
    # pylint: disable=unused-argument, too-many-statements
    def operations_map(
        self,
    ) -> dict[str, OpAndTargetToInstrCallable]:
        """Map Eka operations to PennyLane instructions. The map return a sequence of
        PennyLane instructions, which are specified by name, wires, classical register
        label, and whether it is a measurement or classically controlled operation."""

        def _quantum_gate_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            instructions = []
            if not isinstance(op, list):
                op = [op]
            for o in op:
                if o.startswith(f"{self.import_prefix}") or o.startswith("catalyst_"):
                    prefix = ""
                else:
                    prefix = self.import_prefix
                instructions.append(f"{prefix}{o}({q_targets})")
            return "\n".join(instructions)

        def _measurement_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            res = ""
            if not isinstance(op, list):
                op = [op]
            for o in op[:-1]:
                res += f"{self.import_prefix}{o}({str(q_targets[0])})\n"
            res += f"{desc} = {op[-1]}({str(q_targets[0])})"
            return res

        def _reset_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            """Handle reset operations by calling measure with reset=True, followed by
            any gates."""
            instructions = []
            if not isinstance(op, list):
                op = [op]

            # The first operation should be the measure with reset=True
            measure_func = op[0]
            instructions.append(f"{measure_func}({str(q_targets[0])}, reset=True)")

            # Apply any subsequent gates
            for o in op[1:]:
                if o.startswith(f"{self.import_prefix}") or o.startswith("catalyst_"):
                    prefix = ""
                else:
                    prefix = self.import_prefix
                instructions.append(f"{prefix}{o}({q_targets})")

            return "\n".join(instructions)

        def _control_flow_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            if_block_callable = "call_if_true"
            else_block_callable = "call_else"
            match op:
                case "classical_if":
                    if_instructions = f"def {if_block_callable}():"
                    return if_instructions
                case "classical_else":
                    else_instructions = f"def {else_block_callable}():"
                    return else_instructions
                case "end_if":
                    cond, is_else_present = tuple(
                        desc.split(self.separator_for_else_in_condition)
                    )
                    apply_cond = []
                    apply_cond.append(
                        f"{self.import_prefix}cond({cond}, {if_block_callable})()"
                    )
                    if is_else_present == "True":
                        apply_cond.append(
                            f"{self.import_prefix}cond(~({cond}), "
                            f"{else_block_callable})()"
                        )

                    return "\n".join(apply_cond)
                case _:
                    raise NotImplementedError(
                        f"Control flow operation '{op}' is not supported by PennyLane."
                    )

        def _bool_logic_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str,
            desc: str = "",
        ):
            condition = ""
            match op:
                case BoolOp.MATCH:
                    condition = f"{c_targets[0]} == 1"
                case BoolOp.NOT:
                    condition = f"~({c_targets[0]} == 1)"
                case BoolOp.AND:
                    condition = f"({' == 1) & ('.join(c_targets)} == 1)"
                case BoolOp.OR:
                    condition = f"({' == 1) | ('.join(c_targets)} == 1)"
                case BoolOp.XOR:
                    condition = f"({' == 1) ^ ('.join(c_targets)} == 1)"
                case BoolOp.NAND:
                    condition = f"~ (({' == 1) & ('.join(c_targets)} == 1)) "
                case BoolOp.NOR:
                    condition = f"~ (({' == 1) | ('.join(c_targets)} == 1))"
                case _:
                    condition = ValueError(f"Unsupported BoolOp '{op}'.")
            return condition

        def _utils_op(
            q_targets: list[int],
            c_targets: list[str],
            op: str | list[str],
            desc: str = "",
        ) -> str:
            if op == "comment":
                return f"# {desc}"
            raise ValueError(f"Unsupported utils operation: {op}")

        # Map operation types to their corresponding function
        op_type_handlers = {
            OpType.SINGLE_QUBIT: _quantum_gate_op,
            OpType.TWO_QUBIT: _quantum_gate_op,
            OpType.MEASUREMENT: _measurement_op,
            OpType.RESET: _reset_op,
            OpType.CONTROL_FLOW: _control_flow_op,
            OpType.UTILS: _utils_op,
            OpType.BOOL_LOGIC: _bool_logic_op,
        }

        measure_op = (
            "catalyst_measure" if self.is_catalyst else f"{self.import_prefix}measure"
        )
        # PennyLane implements resets via measure(qubit, reset=True)
        # We use a separate variable for semantic clarity, though it points to the same
        # function
        reset_op = measure_op

        eka_to_pennylane_ops = {
            "i": "Identity",
            "x": "PauliX",
            "y": "PauliY",
            "z": "PauliZ",
            "t": "T",
            "h": "Hadamard",
            "phase": "S",
            "phaseinv": f"adjoint({self.import_prefix}S)",
            "cnot": "CNOT",
            "cy": "CY",
            "cz": "CZ",
            "cx": "CNOT",
            "swap": "SWAP",
            "reset": reset_op,
            "reset_0": reset_op,
            "reset_1": [reset_op, "PauliX"],
            "reset_+": [reset_op, "Hadamard"],
            "reset_-": [reset_op, "PauliX", "Hadamard"],
            "reset_+i": [reset_op, "Hadamard", "S"],
            "reset_-i": [
                reset_op,
                "PauliX",
                "Hadamard",
                "S",
            ],
            "measurement": measure_op,
            "measure_z": measure_op,
            "measure_x": ["Hadamard", measure_op],
            "measure_y": [f"adjoint({self.import_prefix}S)", "Hadamard", measure_op],
        }
        return {
            op.name: partial(
                op_type_handlers[op.op_type], op=eka_to_pennylane_ops[op.name]
            )
            for op in USUAL_QUANTUM_GATES | NONCLIFFORD_GATES_SIGNATURE
        } | {
            op.name: partial(op_type_handlers[op.op_type], op=op.name)
            for op in CONTROL_FLOW_OP_SIGNATURE
            | BOOL_LOGIC_OP_SIGNATURE
            | UTILS_SIGNATURE
        }

    def convert_circuit(
        self, input_circuit: Circuit
    ) -> tuple[str, dict[str, int], dict[str, str]]:
        """Convert an Eka Circuit to PennyLane code string along with quantum and
        classical channel maps.

        Returns
        -------
            A tuple containing:

            - The PennyLane code string representing the circuit.
            - A dictionary mapping Eka quantum channel IDs to PennyLane wire
              indices.
            - A dictionary mapping Eka classical channel IDs to PennyLane
              measurement keys.

        """
        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit")

        pennylane_lines = []

        pennylane_lines.append("# PennyLane Program Generated from Eka")
        if (
            not input_circuit.circuit and input_circuit.name not in self.operations_map
        ) or not input_circuit.channels:
            return "# empty input", {}, {}

        init_instr, q_register, c_register = self.emit_init_instructions(input_circuit)
        pennylane_lines.append("# Initialize Wires from Eka quantum channels Ids")
        pennylane_lines.extend(init_instr.splitlines())

        pennylane_lines.append("")

        circuit_instr = self.emit_circuit_program(
            input_circuit,
            q_register,
            c_register,
        )
        pennylane_lines.extend(circuit_instr.splitlines())

        return "\n".join(pennylane_lines), q_register, c_register

    @staticmethod
    def parse_target_run_outcome(outcome: PennyLaneResult) -> dict[str, list[int]]:
        """Parse the PennyLane/catalyst run outcome into a dictionary mapping the Eka
        classical channels labels to the measurement outcomes.
        """
        result: dict[str, list[int]] = {}
        for key, val in outcome.items():
            result[key] = [int(v) for v in np.ravel(val)]
        return result

    def emit_init_instructions(
        self, circuit: Circuit
    ) -> tuple[str, dict[str, int], dict[str, str]]:
        """Provide a mapping from eka channel label to initialized PennyLane
        wire index."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        q_channels = sorted(
            [chan for chan in circuit.channels if chan.is_quantum()],
            key=lambda c: c.label,
        )
        q_register = {q_chan.id: n for n, q_chan in enumerate(q_channels)}
        init_dict = {q_chan.id: 1 for n, q_chan in enumerate(q_channels)}
        c_register = {
            c_chan.id: f'measurements["{c_chan.label}"]'
            for c_chan in circuit.channels
            if c_chan.is_classical()
        }

        instructions = (
            f"measurements = {{}}\n{self.import_prefix}registers({init_dict})"
        )

        return (instructions, q_register, c_register)

    def emit_leaf_circuit_instruction(
        self,
        circuit: Circuit,
        quantum_channel_map: dict[str, int],
        classical_channel_map: dict[str, str],
    ) -> str:
        """Provide the python code (as a string) to emit an Eka instruction in the
        target language."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if circuit.circuit and len(circuit.circuit) != 0:
            raise ValueError("The circuit must be an leaf circuit with channels.")

        q_targets = [
            quantum_channel_map[q_chan.id]
            for q_chan in circuit.channels
            if q_chan.is_quantum()
        ]

        # Get classical targets
        c_targets = [
            classical_channel_map[c_chan.id]
            for c_chan in circuit.channels
            if c_chan.is_classical()
        ]

        self._validate_ops_args(circuit.name, len(q_targets), len(c_targets))

        if self.op_by_eka_name[circuit.name].op_type == OpType.MEASUREMENT:
            desc = c_targets[0] if c_targets else ""
        else:
            desc = circuit.description

        return self.operations_map[circuit.name](
            q_targets=q_targets, c_targets=c_targets, desc=desc
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from uuid import uuid4

from loom.eka.operations import (
    Merge,
    Split,
    Grow,
    Shrink,
    MeasureBlockSyndromes,
    LogicalMeasurement,
    ConditionalLogicalZ,
)
from loom.eka.utilities import Direction, Orientation
from loom.interpreter.applicator import measureblocksyndromes, conditional_logical_pauli
from loom.interpreter.interpretation_step import InterpretationStep

from .merge import merge
from .split import split
from .grow import grow
from .shrink import shrink
from ..code_factory import RotatedSurfaceCode
from ..operations import AuxCNOT


def auxcnot(
    interpretation_step: InterpretationStep,
    operation: AuxCNOT,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply an auxiliary CNOT operation using a Grow - Shrink - Merge - Split approach.

    The algorithm is the following:
    - A) Begin AuxCNOT composite operation session
    - B) Grow control block
    - C) Measure syndromes of grown_control and target blocks
    - D) Split grown_control into control and auxiliary blocks
    - E) Measure syndromes of control, auxiliary and target blocks
    - F) Merge auxiliary and target blocks
    - G) Measure syndromes of control and merged_target blocks
    - H) Apply ConditionalLogicalZ conditioned on joint measurement
    - I) Shrink the new target block
    - J) End AuxCNOT composite operation session and append circuit

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    operation : AuxCNOT
        The CNOT operation to apply.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after applying the auxiliary CNOT operation.
    """

    c_block = interpretation_step.get_block(operation.input_blocks_name[0])
    t_block = interpretation_step.get_block(operation.input_blocks_name[1])

    # A) Begin AuxCNOT composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"auxcnot between {c_block.unique_label} and {t_block.unique_label}"
        ),
    )

    auxcnot_consistency_check(c_block, t_block)

    grow_direction, shrink_direction = get_grow_shrink_directions(
        control=c_block,
        target=t_block,
    )

    # B) Grow control block
    interpretation_step = auxcnot_grow_control(
        interpretation_step=interpretation_step,
        control=c_block,
        target=t_block,
        grow_direction=grow_direction,
        same_timeslice=False,  # Prevent the same timeslice flag from being set
        debug_mode=debug_mode,
    )

    # Get the grown control block
    grown_control = interpretation_step.get_block(c_block.unique_label)
    aux_unique_label = f"{grown_control.unique_label}_aux"
    # Ensure the auxiliary block has a unique label
    if aux_unique_label in [
        block.unique_label for block in interpretation_step.get_blocks_at_index(-1)
    ]:
        aux_unique_label = str(uuid4())

    # C) Measure syndromes of grown_control and target blocks
    for block, internal_same_timeslice in ((grown_control, False), (t_block, True)):
        interpretation_step = measureblocksyndromes(
            interpretation_step=interpretation_step,
            operation=MeasureBlockSyndromes(block.unique_label, min(block.size)),
            same_timeslice=internal_same_timeslice,
            debug_mode=debug_mode,
        )

    # D) Split grown_control into control and auxiliary blocks
    interpretation_step = auxcnot_split_control(
        interpretation_step=interpretation_step,
        aux_unique_label=aux_unique_label,
        initial_control=c_block,
        initial_target=t_block,
        grown_control=grown_control,
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # Get the new control and auxiliary blocks
    new_control = interpretation_step.get_block(c_block.unique_label)
    aux_block = interpretation_step.get_block(aux_unique_label)

    # E) Measure syndromes of control, auxiliary and target blocks
    for block, internal_same_timeslice in (
        (new_control, False),
        (aux_block, True),
        (t_block, True),
    ):
        interpretation_step = measureblocksyndromes(
            interpretation_step=interpretation_step,
            operation=MeasureBlockSyndromes(block.unique_label, 1),
            same_timeslice=internal_same_timeslice,
            debug_mode=debug_mode,
        )

    # F) Merge auxiliary and target blocks
    interpretation_step = auxcnot_merge_aux_target(
        interpretation_step=interpretation_step,
        aux=aux_block,
        target=t_block,
        same_timeslice=False,  # Prevent the same timeslice flag from being set
        debug_mode=debug_mode,
    )

    # Get the merged target block
    merged_target = interpretation_step.get_block(t_block.unique_label)

    # G) Measure syndromes of control and merged_target blocks
    for block, internal_same_timeslice, n_cycles in (
        (new_control, True, 1),
        (new_control, False, min(new_control.size) - 1),
        (merged_target, True, min(merged_target.size) - 1),
    ):
        interpretation_step = measureblocksyndromes(
            interpretation_step=interpretation_step,
            operation=MeasureBlockSyndromes(block.unique_label, n_cycles),
            same_timeslice=internal_same_timeslice,
            debug_mode=debug_mode,
        )

    # H) Apply ConditionalLogicalZ conditioned on joint measurement
    interpretation_step = auxcnot_conditional_logical_z(
        interpretation_step=interpretation_step,
        new_control=new_control,
        initial_target=t_block,
        aux_block=aux_block,
        new_target=merged_target,
        shrink_direction=shrink_direction,
        same_timeslice=True,
        debug_mode=debug_mode,
    )

    # I) Shrink the new target block
    interpretation_step = auxcnot_shrink_target(
        interpretation_step=interpretation_step,
        initial_target=t_block,
        merged_target=merged_target,
        shrink_direction=shrink_direction,
        same_timeslice=False,  # Prevent the same timeslice flag from being set
        debug_mode=debug_mode,
    )

    # J) End AuxCNOT composite operation session and append circuit
    auxcnot_circ = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(auxcnot_circ, same_timeslice)

    # Return the final step
    return interpretation_step


def auxcnot_consistency_check(
    c_block: RotatedSurfaceCode,
    t_block: RotatedSurfaceCode,
):
    """
    Perform multiple checks to ensure that the auxiliary CNOT operation can be applied.
    The blocks must be of the correct type, have the same size, the same boundary
    orientations and be in a specific configuration.

    The configuration is the following, the upper left corners of the two blocks must
    satisfy the following relations:

    - ``|t_block.upper_left_qubit[0] - c_block.upper_left_qubit[0]| = c_block.size[0]``
    - ``|t_block.upper_left_qubit[1] - c_block.upper_left_qubit[1]| = c_block.size[1]``

    Parameters
    ----------
    c_block : RotatedSurfaceCode
        Control block for the auxiliary CNOT operation.
    t_block : RotatedSurfaceCode
        Target block for the auxiliary CNOT operation.
    """

    if not isinstance(c_block, RotatedSurfaceCode) or not isinstance(
        t_block, RotatedSurfaceCode
    ):
        raise TypeError(
            "The auxcnot operation is only supported for RotatedSurfaceCode blocks. "
            f"{set(type(block) for block in [c_block, t_block])} types "
            "were given."
        )

    # NOTE: this is a sub-case, can be extended later
    if c_block.size != t_block.size:
        raise ValueError(
            f"The blocks must have the same size to perform the auxiliary CNOT "
            f"operation. The sizes of the blocks are {c_block.size}, {t_block.size}."
        )

    # Check that the blocks have the correct boundary orientations
    if c_block.x_boundary != t_block.x_boundary:
        raise ValueError(
            "The blocks must have the same boundary orientations to perform the "
            "auxiliary CNOT operation. The X boundary orientations are "
            f"{c_block.x_boundary} and {t_block.x_boundary}."
        )

    # Check that the blocks are in the correct configuration
    if (
        abs(t_block.upper_left_qubit[0] - c_block.upper_left_qubit[0])
        != c_block.size[0] + 1  # +1 because the auxiliary qubit is one row/column away
        or abs(t_block.upper_left_qubit[1] - c_block.upper_left_qubit[1])
        != c_block.size[1] + 1  # +1 because the auxiliary qubit is one row/column away
    ):
        raise ValueError(
            "The blocks are not in the correct configuration for the auxiliary CNOT "
            "operation. The upper left corners of the blocks must satisfy the "
            "following relations: \n"
            "|t_block.upper_left_qubit[0] - c_block.upper_left_qubit[0]| = "
            "c_block.size[0] + 1, |t_block.upper_left_qubit[1] - "
            "c_block.upper_left_qubit[1]| = c_block.size[1] + 1"
            f"\nGot |{t_block.upper_left_qubit[0]} - {c_block.upper_left_qubit[0]}| = "
            f"{c_block.size[0] + 1}, |{t_block.upper_left_qubit[1]} - "
            f"{c_block.upper_left_qubit[1]}| = {c_block.size[1] + 1} instead."
        )


def get_grow_shrink_directions(
    control: RotatedSurfaceCode,
    target: RotatedSurfaceCode,
) -> tuple[Direction, Direction]:
    """
    Get the grow and shrink directions for the auxiliary CNOT operation.
    The grow direction is the direction of the X operator of the control block, and
    the shrink direction is the direction normal to the X operator of the target block.

    Parameters
    ----------
    control : RotatedSurfaceCode
        Control block for the auxiliary CNOT operation.
    target : RotatedSurfaceCode
        Target block for the auxiliary CNOT operation.

    Returns
    -------
    tuple[Direction, Direction]
        The grow and shrink directions.
    """
    upper_left_qubits_vector = (
        control.upper_left_qubit[0] - target.upper_left_qubit[0],
        control.upper_left_qubit[1] - target.upper_left_qubit[1],
    )
    match upper_left_qubits_vector:
        case (x, y) if x < 0 and y < 0:
            # The control block is to the left and above the target block
            grow_direction = (
                Direction.RIGHT
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.BOTTOM
            )
            shrink_direction = (
                Direction.TOP
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.LEFT
            )
        case (x, y) if x < 0 < y:
            # The control block is to the left and below the target block
            grow_direction = (
                Direction.RIGHT
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.TOP
            )
            shrink_direction = (
                Direction.BOTTOM
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.LEFT
            )
        case (x, y) if x > 0 > y:
            # The control block is to the right and above the target block
            grow_direction = (
                Direction.LEFT
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.BOTTOM
            )
            shrink_direction = (
                Direction.TOP
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.RIGHT
            )
        case (x, y) if x > 0 and y > 0:
            # The control block is to the right and below the target block
            grow_direction = (
                Direction.LEFT
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.TOP
            )
            shrink_direction = (
                Direction.BOTTOM
                if control.x_boundary == Orientation.HORIZONTAL
                else Direction.RIGHT
            )
        case _:
            raise ValueError(
                "The control block is not in the correct position relative to the "
                "target block."
            )
    return grow_direction, shrink_direction


def auxcnot_grow_control(
    interpretation_step: InterpretationStep,
    control: RotatedSurfaceCode,
    target: RotatedSurfaceCode,
    grow_direction: Direction,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Grow the control block for the auxiliary CNOT operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    control : RotatedSurfaceCode
        Control block for the auxiliary CNOT operation.
    target : RotatedSurfaceCode
        Target block for the auxiliary CNOT operation.
    grow_direction : Direction
        Direction in which to grow the control block.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after growing the control block.
    """
    # Grow the control block in the orientation of its X operator
    grow_length = (
        target.size[0] + 1
        if control.x_boundary == Orientation.HORIZONTAL
        else target.size[1] + 1
    )

    grow_op = Grow(
        input_block_name=control.unique_label,
        direction=grow_direction,
        length=grow_length,
    )
    interpretation_step = grow(
        interpretation_step=interpretation_step,
        operation=grow_op,
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )
    return interpretation_step


def auxcnot_split_control(
    interpretation_step: InterpretationStep,
    aux_unique_label: str,
    initial_control: RotatedSurfaceCode,
    initial_target: RotatedSurfaceCode,
    grown_control: RotatedSurfaceCode,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Split the grown control block into the control and auxiliary blocks.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    aux_unique_label : str
        Unique label for the auxiliary block that will be created after the split.
    initial_control : RotatedSurfaceCode
        Initial control block before it was grown.
    initial_target : RotatedSurfaceCode
        Initial target block before the auxiliary CNOT operation.
    grown_control : RotatedSurfaceCode
        Control block for the auxiliary CNOT operation that was just grown.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after splitting the grown control block.
    """

    # Split the grown control block into the control and auxiliary blocks
    split_orientation = grown_control.x_boundary.perpendicular()

    # Ensure the order of naming is right, if the initial upper-left qubit of the
    # control block is the same as the grown control block, then the split control will
    # also inherit that qubit and the first name in the tuple.
    # The position of the split also depends on the size of the block that is located on
    # the top-left
    if initial_control.upper_left_qubit == grown_control.upper_left_qubit:
        split_blocks_name = (grown_control.unique_label, aux_unique_label)
        split_position = (
            initial_control.size[0]
            if split_orientation == Orientation.VERTICAL
            else initial_control.size[1]
        )
    else:
        split_blocks_name = (aux_unique_label, grown_control.unique_label)
        split_position = (
            initial_target.size[0]
            if split_orientation == Orientation.VERTICAL
            else initial_target.size[1]
        )

    split_op = Split(
        input_block_name=grown_control.unique_label,
        output_blocks_name=split_blocks_name,
        orientation=split_orientation,
        split_position=split_position,
    )

    interpretation_step = split(
        interpretation_step=interpretation_step,
        operation=split_op,
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )

    return interpretation_step


def auxcnot_merge_aux_target(
    interpretation_step: InterpretationStep,
    aux: RotatedSurfaceCode,
    target: RotatedSurfaceCode,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Merge the auxiliary block with the target block.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    aux : RotatedSurfaceCode
        Auxiliary block for the auxiliary CNOT operation.
    target : RotatedSurfaceCode
        Target block for the auxiliary CNOT operation.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after merging the auxiliary and target blocks.
    """
    # Merge the auxiliary block with the target block
    merge_op = Merge(
        input_blocks_name=(aux.unique_label, target.unique_label),
        output_block_name=target.unique_label,
    )

    interpretation_step = merge(
        interpretation_step=interpretation_step,
        operation=merge_op,
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )

    return interpretation_step


def auxcnot_shrink_target(
    interpretation_step: InterpretationStep,
    initial_target: RotatedSurfaceCode,
    merged_target: RotatedSurfaceCode,
    shrink_direction: Direction,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Shrink the target block after merging with the auxiliary block.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    initial_target : RotatedSurfaceCode
        Initial target block for the auxiliary CNOT operation.
    merged_target : RotatedSurfaceCode
        Merged target block after merging with the auxiliary block.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after shrinking the target block.
    """
    # Shrink the target block in the direction normal to its X operator
    shrink_length = (
        merged_target.size[0] - initial_target.size[0]
        if merged_target.x_boundary == Orientation.VERTICAL
        else merged_target.size[1] - initial_target.size[1]
    )

    shrink_op = Shrink(
        input_block_name=merged_target.unique_label,
        direction=shrink_direction,
        length=shrink_length,
    )

    interpretation_step = shrink(
        interpretation_step=interpretation_step,
        operation=shrink_op,
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )

    return interpretation_step


def auxcnot_conditional_logical_z(
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    interpretation_step: InterpretationStep,
    new_control: RotatedSurfaceCode,
    initial_target: RotatedSurfaceCode,
    aux_block: RotatedSurfaceCode,
    new_target: RotatedSurfaceCode,
    shrink_direction: Direction,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Applies ConditionalLogicalZ to the blocks conditioned on the value of the joint
    measurement obtained from Merge.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Input interpretation step.
    new_control : RotatedSurfaceCode
        New control block after the AuxCNOT operations.
    initial_target : RotatedSurfaceCode
        Target block before the the AuxCNOT operations.
    aux_block : RotatedSurfaceCode
        Auxiliary block for the AuxCNOT operation.
    new_target : RotatedSurfaceCode
        New target block after the AuxCNOT operations.
    shrink_direction : Direction
        Direction in which the target block was shrunk.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the input.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after applying the LogicalZ
    """

    log_meas = LogicalMeasurement(
        (initial_target.unique_label, aux_block.unique_label), "XX"
    )

    log_op = ConditionalLogicalZ(new_control.unique_label, condition=log_meas)

    interpretation_step = conditional_logical_pauli(
        interpretation_step=interpretation_step,
        operation=log_op,
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )

    # Additional correction needed if merged_target inherits the auxiliary block's
    # logical X operator, which happens when shrink_direction is "TOP" or "LEFT"

    if shrink_direction in {Direction.TOP, Direction.LEFT}:
        log_op = ConditionalLogicalZ(new_target.unique_label, condition=log_meas)

        interpretation_step = conditional_logical_pauli(
            interpretation_step=interpretation_step,
            operation=log_op,
            same_timeslice=same_timeslice,
            debug_mode=debug_mode,
        )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit
from loom.eka.operations import ResetAllAncillaQubits

from ..interpretation_step import InterpretationStep


def reset_all_ancilla_qubits(
    interpretation_step: InterpretationStep,
    operation: ResetAllAncillaQubits,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Resets all ancilla qubits of a block to a specific SingleQubitPauliEigenstate.

    NOTE: Initializing a Y state may come with some caveats, as the implementation of
    the initialization may not be fault-tolerant for some codes. For example, in the
    case of the Rotated Surface Code, initializing a Y state may require distillation
    for it to be fault-tolerant.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks whose ancilla qubits need to be reset.
    operation : ResetAllAncillaQubits
        Reset ancilla operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the reset ancilla operation.
    """

    # Get the block
    block = interpretation_step.get_block(operation.input_block_name)

    # Create a circuit that resets the data qubits to the given state
    reset_circuit = Circuit(
        name=f"reset all ancilla qubits of block {block.unique_label} to "
        f"|{operation.state.value}>",
        circuit=[
            # Reset the data qubits on the same time step
            [
                Circuit(
                    f"reset_{operation.state.value}",
                    channels=interpretation_step.get_channel_MUT(q, "quantum"),
                )
                for q in block.ancilla_qubits
            ]
        ],
    )
    # Add the reset circuit to the interpretation step
    interpretation_step.append_circuit_MUT(reset_circuit, same_timeslice)

    # Return the new interpretation step
    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np

from .stab_array import StabArray, reindex_stabarray, stabarray_standard_form


def find_logical_operator_set(stabarr: StabArray) -> tuple[StabArray, StabArray]:
    """Given a stabilizer array of a code, find a set of X and Z logical operators.
    The Z logical operators shall contain only Z operators while the X logical operators
    may contain both X and Z operators. They may not be in their simplified form.

    This algorithm follows the notation and algorithm described in:
    Nielsen, M. A., & Chuang, I. L. (2011). (p.470-471) Quantum Computation and Quantum
    Information: 10th Anniversary Edition. Cambridge University Press,

    How it works (ignoring signs):

        1. The stabilizer array is first put into standard form. The standard form is of
        the form:

            (I, A1, A2 | B, 0, C)
            (0,  0,  0 | D, I, E)

        Note that the stabilizer array has to be reindexed to the standard form.

        2. The Z logical operators are then given by:

            (0, 0, 0 | A2^T, 0, I)

        and the X logical operators are given by:

            (0, E^T, I | C^T , 0, 0)

        3. Reindex the logical operators to the original qubit order.

    Parameters
    ----------
    stabarr : StabArray

    Returns
    -------
    tuple[StabArray, StabArray]
        A tuple containing the X and Z logical operator StabArray objects.
    """
    stabarray_std_form, r, inv_qbit_map = stabarray_standard_form(stabarr)

    n = stabarr.nqubits  # Number of qubits
    k = n - stabarray_std_form.nstabs  # Number of logical qubits

    # Find the auxiliary arrays A2, E and C
    a2_array = stabarray_std_form.array[:r, n - k : n]
    e_array = stabarray_std_form.array[r:, -k - 1 : -1]  # -1 to exclude the sign column
    c_array = stabarray_std_form.array[:r, -k - 1 : -1]  # -1 to exclude the sign column

    # Construct the logical operator arrays and cast the data to the correct type
    z_logop_arr = np.hstack(
        [
            np.zeros((k, stabarr.nqubits), dtype=StabArray.DTYPE),
            a2_array.transpose(),
            np.zeros((k, n - k - r), dtype=StabArray.DTYPE),
            np.eye(k, dtype=StabArray.DTYPE),
            np.zeros((k, 1), dtype=StabArray.DTYPE),  # the sign column
        ],
        dtype=StabArray.DTYPE,
    )

    x_logop_arr = np.hstack(
        [
            np.zeros((k, r), dtype=StabArray.DTYPE),
            e_array.transpose(),
            np.eye(k, dtype=StabArray.DTYPE),
            c_array.transpose(),
            np.zeros((k, n - r), dtype=StabArray.DTYPE),
            np.zeros((k, 1), dtype=StabArray.DTYPE),  # the sign column
        ],
        dtype=StabArray.DTYPE,
    )

    # Cast the logical operator arrays to StabArray objects and reindex them to the
    # original
    # NOTE: We bypass the validation of the StabArray objects since we know that the
    # logical operator arrays are valid
    z_log_ops_stabarr = reindex_stabarray(
        StabArray(z_logop_arr, validated=True), inv_qbit_map
    )

    x_log_ops_stabarr = reindex_stabarray(
        StabArray(x_logop_arr, validated=True), inv_qbit_map
    )

    return x_log_ops_stabarr, z_log_ops_stabarr



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import json
from dataclasses import asdict


def findall(in_list, value):
    """
    Find all the indices of a value in a list.
    """
    return [i for i, x in enumerate(in_list) if x == value]


def apply_to_nested(nested_list, func, param=None):
    """
    Apply a function to all elements of a nested list.
    """
    if isinstance(nested_list, list):
        return [apply_to_nested(elem, func, param) for elem in nested_list]
    if param is not None:
        return func(nested_list, param)
    return func(nested_list)  # if neither of the above two conditions is satisfied


def dumps(data) -> str:
    """
    Return the pydantic dataclass as a JSON string. If the dataclass defines its own
    custom dumps() method, the custom method is used. Otherwise, the dataclass object is
    converted to a dictionary using the `asdict` method of the `dataclasses` module and
    then converted to a JSON string.

    Parameters
    ----------
    data : dataclass
        Pydantic dataclass to be saved as JSON

    Returns
    -------
    str
        JSON string representation of the Eka
    """

    # Check if the dataclass defines its own custom dumps() method and call it if it is
    # defined
    if hasattr(data, "asdict") and callable(data.asdict):
        return json.dumps(data.asdict())

    # Otherwise use default methods
    return json.dumps(asdict(data))


def dump(data, file: str):
    """
    Write a pydantic dataclass as JSON to a file.

    Parameters
    ----------
    data : dataclass
        Pydantic dataclass to be saved as JSON
    file : str
        Filename to write to
    """

    with open(file, "w", encoding="utf-8") as f:
        f.write(dumps(data))


def loads(cls, data_json: str):
    """
    Load the a pydantic dataclass from a JSON string. If the dataclass defines its own
    custom loads() method, the custom method is used, giving the loaded dictionary as
    input.

    Parameters
    ----------
    cls : dataclass
        The dataclass to be loaded from the JSON string
    data_json : str
        Dataclass saved as JSON string

    Returns
    -------
    An instance of cls containing the data from the JSON string
    """
    if hasattr(cls, "fromdict") and callable(cls.fromdict):
        return cls.fromdict(json.loads(data_json))

    return cls(**json.loads(data_json))


def load(cls, file: str):
    """
    Load a pydantic dataclass from a JSON file.

    Parameters
    ----------
    cls : dataclass
        The dataclass to be loaded from the JSON file
    file : str
        Filename to read from

    Returns
    -------
    An instance of cls containing the data from the JSON file
    """

    with open(file, "w", encoding="utf-8") as f:
        return cls(**json.loads(f.read()))



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from loom.eka import ChannelType, Circuit, PauliOperator, Stabilizer
from loom.eka.operations import Shrink
from loom.eka.utilities import Direction
from loom.interpreter import InterpretationStep
from loom.interpreter.utilities import Cbit

from ..code_factory import RepetitionCode


# pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals
def shrink_consistency_check(
    interpretation_step: InterpretationStep, operation: Shrink
) -> RepetitionCode:
    """
    Check the consistency of the shrink operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        InterpretationStep containing the block to shrink.
    operation : Shrink
        Shrink operation description.

    Returns
    -------
    RepetitionCode
        Block to shrink.
    """
    # Extract Block
    block = interpretation_step.get_block(operation.input_block_name)

    # Check valid Block type
    if not isinstance(block, RepetitionCode):
        raise ValueError(
            f"This shrink operation is not supported for {type(block)} blocks."
        )

    # Check shrink size will leave more than one data qubit in the block
    if operation.length >= (len(block.data_qubits) - 1):
        raise ValueError("Shrink size is too large.")

    # Check shrink is applied in the correct direction
    if operation.direction in [Direction.BOTTOM, Direction.TOP]:
        raise ValueError(
            "Repetition code does not support "
            f"shrinking in the {operation.direction} direction."
        )

    return block


def get_qubits_to_measure(
    block: RepetitionCode, direction: Direction, length: int
) -> list[tuple[int, int]]:
    """Find the qubits to measure during the shrink operation.

    Parameters
    ----------
    block : RepetitionCode
        Block to shrink.
    direction : Direction
        Direction of the shrink operation.
    length : int
        Length of the shrink operation.

    Returns
    -------
    list[tuple[int, int]]
        List containing the qubits to measure.
    """

    # Extract qubit in the boundary to be shrunk
    boundary_qubit = block.boundary_qubits(direction=direction)

    # Generate a list of vectors shifting the new left boundary qubit after shrink
    match direction:
        case Direction.LEFT:
            shift_vectors = [(i, 0) for i in range(length)]
        case Direction.RIGHT:
            shift_vectors = [(-i, 0) for i in range(length)]

    # Extract qubits to measure, using the boundary as a reference
    qubits_to_measure = [
        tuple(coord1 + coord2 for coord1, coord2 in zip(boundary_qubit, shift_vect))
        for shift_vect in shift_vectors
    ]

    return qubits_to_measure  # type: ignore


def find_shrink_circuit(
    interpretation_step: InterpretationStep,
    check_type: str,
    qubits_to_measure: list[tuple[int, int]],
    circuit_name: str,
) -> tuple[Circuit, list[Cbit]]:
    """Generate a circuit to measure the qubits to be shrunk. Qubits are measured in
    the Z basis. If the check type is Z, the qubits are measured directly, else they
    are applied a Hadamard gate before measurement.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to shrink.
    check_type : str
        Type of the stabilizer operators.
    qubits_to_measure : list[tuple[int,int]]
        List of qubits to measure.
    circuit_name : str
        Name of the circuit.

    Returns
    -------
    tuple[Circuit, list[Cbit]]
        Tuple containing the shrink circuit and the classical bits associated with the
        measured qubits.
    """

    # Generate fresh classical bits and extract the channels
    cbits = [interpretation_step.get_new_cbit_MUT(f"c_{q}") for q in qubits_to_measure]

    cbit_channels = [
        interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type=ChannelType.CLASSICAL
        )
        for cbit in cbits
    ]

    # Create sequence of measurements for every measured data qubit
    measure_circuit_seq = [
        [
            Circuit(
                "Measurement",
                channels=[interpretation_step.get_channel_MUT(q), cbc],
            )
            for q, cbc in zip(qubits_to_measure, cbit_channels, strict=True)
        ]
    ]

    if check_type == "X":
        # If phaseflip code qubits can already be measured
        shrink_circuit_list = measure_circuit_seq
    else:
        # If bitflip code qubits need to be Hadamard transformed before measurement
        basis_change_circuit_seq = [
            [
                Circuit("H", channels=[interpretation_step.get_channel_MUT(qb)])
                for qb in qubits_to_measure
            ]
        ]

        # Add basis change before measurement
        shrink_circuit_list = basis_change_circuit_seq + measure_circuit_seq

    # Construct shrink circuit
    shrink_circuit = Circuit(
        name=circuit_name,
        circuit=shrink_circuit_list,
    )

    return shrink_circuit, cbits


def find_new_stabilizers(
    block: RepetitionCode, qubits_to_measure: list[tuple[int, int]]
) -> list[Stabilizer]:
    """Find the new set of stabilizers after the shrink operation.
    We also compute the uuids of the stabilizers that need to be removed, if the block
    shrunk from the left, as they will be used to update one of the logical operators.

    Parameters
    ----------
    block : RepetitionCode
        Block to shrink.
    qubits_to_measure : list[tuple[int,int]]
        List of qubits to measure.

    Returns
    -------
    new_stabs : list[Stabilizer]
        New set of stabilizers after the shrink operation.
    """

    stabs_to_remove = [
        stab
        for stab in block.stabilizers
        if any(qb in qubits_to_measure for qb in stab.data_qubits)
    ]

    # Combine the stabilizers to get the new set of stabilizers
    new_stabs = list(set(block.stabilizers) - set(stabs_to_remove))

    return new_stabs


def get_logical_operator_and_updates(
    interpretation_step: InterpretationStep,
    block: RepetitionCode,
    check_type: str,
    is_left: bool,
    qubits_to_measure: list[tuple[int, int]],
    cbits: list[Cbit],
) -> tuple[
    list[list[PauliOperator]],
    list[dict[str, tuple[str, ...]]],
    list[dict[str, tuple[Cbit, ...]]],
]:
    """
    Generate new logical operators and their respective evolution and updates to be
    added to the interpretation step. We treat the two logicals separately in as "long"
    and "short". The long corresponds to the one covering the entire chain, with type
    opposite to the check type. The short corresponds to the one covering the left
    boundary qubit by default, with type equal to the check type.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    block : RepetitionCode
        Block to shrink.
    check_type : str
        Type of the stabilizer operators.
    is_left : bool
        Boolean defining whether shrink operations is performed from right or left.
    qubits_to_measure : list[tuple[int,int]]
        List of qubits to measure.
    cbits : list[Cbit]
        List of classical bits associated with measured qubits.

    Returns
    -------
    tuple[
        list[list[PauliOperator]],
        list[dict[str,tuple[str,...]]],
        list[dict[str,tuple[Cbit,...]]]
    ]
        Tuple containing the new logical operators, their evolution and updates.
    """

    other_check_type = "X" if check_type == "Z" else "Z"

    # Find remaining qubits
    remaining_qubits = list(set(block.data_qubits) - set(qubits_to_measure))

    # Extract old long and short operators
    old_logs = [block.logical_x_operators[0], block.logical_z_operators[0]]
    old_short_log, old_long_log = sorted(old_logs, key=lambda x: len(x.data_qubits))

    # Check if the short logical qubit inside the qubits to measure
    to_shift = old_short_log.data_qubits[0] in qubits_to_measure

    # Define new long logical operator and updates
    new_long_log = PauliOperator(
        pauli=other_check_type * len(remaining_qubits), data_qubits=remaining_qubits
    )
    long_log_updates = {new_long_log.uuid: tuple(cbits)}
    long_log_evolution = {new_long_log.uuid: (old_long_log.uuid,)}

    # Define new short logical operator and updates
    # If the short logical operator is inside the qubits to measure, shift it
    if to_shift:
        selector_function = min if is_left else max
        new_qubit = selector_function(remaining_qubits, key=lambda x: x[0])
        new_short_log, stabs_required = block.get_shifted_equivalent_logical_operator(
            new_qubit
        )
        id_stabs_required = [stab.uuid for stab in stabs_required]
        short_log_evolution = {
            new_short_log.uuid: tuple([old_short_log.uuid] + id_stabs_required)
        }

        # Add to the updates the cbits associated with the measured qubits and the
        # stabilizers required for the shift.
        # These cbits are the last syndromes measured for those stabilizers.
        cbits = interpretation_step.retrieve_cbits_from_stabilizers(
            stabs_required, block
        )

        short_log_updates = {new_short_log.uuid: cbits}

    else:
        new_short_log = old_short_log
        short_log_evolution = {}
        short_log_updates = {}

    # New logicals
    new_logs = [[new_long_log], [new_short_log]]

    # Updates and evolution
    log_evolution = [long_log_evolution, short_log_evolution]
    log_updates = [long_log_updates, short_log_updates]

    return new_logs, log_evolution, log_updates


def shrink(
    interpretation_step: InterpretationStep,
    operation: Shrink,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Shrink a Repetition Code chain in the left or right specified direction.

    The algorithm is the following:

    - A.) DATA QUBITS

        - A.1) Find measured data qubits during shrink

    - B.) CIRCUIT

        - B.1) Generate shrink circuit and classical bits
        - B.2) Add circuit to the interpretation step

    - C.) - STABILIZERS

        - C.1) Find new set of stabilizers

    - D.) LOGICAL OPERATORS

        - D.1) Extract logical operators and updates
        - D.2) Update the logical operator history

    - E.) NEW BLOCK AND NEW INTERPRETATION STEP
        - E.1) Create the new block
        - E.2) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to shrink.
    operation : Shrink
        Shrink operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the shrink operation.
    """

    # Check consistency of the shrink operation
    block = shrink_consistency_check(interpretation_step, operation)

    # Get shrink direction
    is_left = operation.direction == Direction.LEFT

    # Extract stabilizer type
    check_type = block.check_type

    # A) DATA QUBITS
    # A.1) Find measured data qubits during shrink
    qubits_to_measure = get_qubits_to_measure(
        block, direction=operation.direction, length=operation.length
    )

    # B) CIRCUIT
    # B.1) Generate shrink circuit and classical bits
    circuit_name = (
        f"shrink {block.unique_label} by {operation.length} from {operation.direction}"
    )
    shrink_circuit, cbits = find_shrink_circuit(
        interpretation_step, check_type, qubits_to_measure, circuit_name
    )

    # B.2) Add circuit to the interpretation step
    interpretation_step.append_circuit_MUT(shrink_circuit, same_timeslice)

    # C) STABILIZERS
    # C.1) Find new set of stabilizers
    new_stabilizers = find_new_stabilizers(block, qubits_to_measure)

    # D) LOGICAL OPERATORS
    # D.1) Extract logical operators and updates
    new_logs, log_evolution, log_updates = get_logical_operator_and_updates(
        interpretation_step,
        block,
        check_type,
        is_left,
        qubits_to_measure,
        cbits,
    )

    # D.2) Update the logical operator history
    ordering = 1 if check_type == "Z" else -1
    new_log_x_ops, new_log_z_ops = new_logs[::ordering]
    logical_x_evolution, logical_z_evolution = log_evolution[::ordering]
    logical_x_updates, logical_z_updates = log_updates[::ordering]

    # Update the logical operator evolution
    interpretation_step.logical_x_evolution.update(logical_x_evolution)
    interpretation_step.logical_z_evolution.update(logical_z_evolution)

    # Update the logical operator updates
    for op_id, measurements in logical_x_updates.items():
        interpretation_step.update_logical_operator_updates_MUT(
            "X",
            op_id,
            measurements,
            True,
        )
    for op_id, measurements in logical_z_updates.items():
        interpretation_step.update_logical_operator_updates_MUT(
            "Z",
            op_id,
            measurements,
            True,
        )

    # E) NEW BLOCK
    # E.1) Create the new block
    new_block = RepetitionCode(
        stabilizers=new_stabilizers,
        logical_x_operators=new_log_x_ops,
        logical_z_operators=new_log_z_ops,
        unique_label=block.unique_label,
        skip_validation=not debug_mode,
    )

    # E.2) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,),
        old_blocks=(block,),
    )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import cached_property, partial
from typing import Any

from pydantic import Field, field_validator

from ..eka import Circuit, Channel, IfElseCircuit
from ..eka.utilities import BoolOp


from .op_signature import (
    CLIFFORD_GATES_SIGNATURE,
    USUAL_QUANTUM_GATES,
    UTILS_SIGNATURE,
    NONCLIFFORD_GATES_SIGNATURE,
    OpSignature,
    OpType,
)
from .converter import Converter, OpAndTargetToInstrCallable

MimiqResult = tuple[Any, dict[Channel, int]]


class EkaToMimiqConverter(Converter[Channel, MimiqResult]):
    """Convert an InterpretationStep to a Mimiq circuit.

    .. code-block:: python

        import mimiqcircuits as mymc
        mimiq_alias = "mymc"
        from loom.executor import EkaToMimiqConverter

        conn = mymc.MimiqConnection()
        conn.connect("username", "pwd")

        mimiq_exec = EkaToMimiqConverter(
            mimiq_import_prefix=f"{mimiq_alias}.", circuit_varname="my_circuit"
        )

        program_str, qreg, creg = mimiq_exec.convert(interpreted_eka)

        p = (
            f"import mimiqcircuits as {mimiq_alias}\\n{program_str}"
        )

        local_ns = {}
        exec(p, {}, local_ns)

        res = local_ns["my_circuit"]

        job = conn.execute(res, algorithm="mps", nsamples=5)
        mcres = conn.get_result(job)

        parsed_outcome = EkaToMimiqConverter.parse_target_run_outcome((mcres, creg))

    Parameters
    ----------
    mimiq_import_prefix : str
        The import alias for Mimiq. defaults to "mc.".
    circuit_varname : str
        The name of the Mimiq circuit variable.
    """

    SUPPORTED_OPERATIONS: frozenset[OpSignature] = USUAL_QUANTUM_GATES | UTILS_SIGNATURE

    ALLOW_ERROR_MODELS: bool = Field(default=False, frozen=True, init=False)

    MIMIQ_CLASSICALLY_CONTROLLED_OPS: frozenset[OpSignature] = frozenset(
        {
            OpSignature(
                name=m,
                op_type=OpType.CUSTOM,
                classical_input=1,
                quantum_input=1,
            )
            for m in {
                f"classical_controlled_{n.name}" for n in CLIFFORD_GATES_SIGNATURE
            }
        }
    )

    mimiq_import_prefix: str = Field(
        default="mc.",
        frozen=True,
        init=True,
        description="The import alias for Mimiq.",
    )

    circuit_varname: str = Field(
        default="circuit",
        frozen=True,
        init=True,
        description="The name of the Mimiq circuit variable.",
    )

    @field_validator("mimiq_import_prefix")
    # pylint: disable=no-self-argument
    def validate_import_prefix(cls, v: str) -> str:
        """Ensure the import prefix ends with a dot if not empty."""
        if v and not v.endswith("."):
            v += "."
        return v

    @cached_property
    # pylint: disable=unused-argument
    def operations_map(self) -> dict[str, OpAndTargetToInstrCallable]:
        """Map of operation signatures to their corresponding Mimiq operations."""

        def _push_string(op: str, targets: list[int]) -> str:
            t = f"{', '.join(map(str, targets))}"
            return f"{self.circuit_varname}.push({self.mimiq_import_prefix}{op}, {t})"

        def _quantum_op(
            q_target: list[int],
            c_target: list[int],
            op: str | list[str],
            desc: str = "",
        ):

            if isinstance(op, list):
                return "\n".join(_push_string(o, q_target + c_target) for o in op)
            return _push_string(op, q_target + c_target)

        def _utils_op(
            q_target: list[int],
            c_target: list[int],
            op: str,
            desc: str = "",
        ):
            if op == "comment":
                return f"# {desc}"
            raise ValueError(f"Unsupported utils operation: {op}")

        def _control_flow_op(
            q_target: list[int],
            c_target: list[int],
            op: str | list[str],
            desc: str = "",
        ):
            op = eka_to_mimiq_ops[op.replace("classical_controlled_", "")]
            cond = f'{self.mimiq_import_prefix}BitString("{desc}")'

            def if_then_op(o):
                return (
                    f"{self.mimiq_import_prefix}IfStatement("
                    f"{self.mimiq_import_prefix}{o}, {cond})"
                )

            targets = f"{', '.join(map(str, q_target+c_target))}"
            if not isinstance(op, list):
                op = [op]
            return "\n".join(
                f"{self.circuit_varname}.push({if_then_op(o)}, {targets})" for o in op
            )

        # Map operation types to their corresponding function
        op_type_handlers = {
            OpType.SINGLE_QUBIT: _quantum_op,
            OpType.TWO_QUBIT: _quantum_op,
            OpType.MEASUREMENT: _quantum_op,
            OpType.RESET: _quantum_op,
            OpType.UTILS: _utils_op,
            OpType.CUSTOM: _control_flow_op,
        }

        eka_to_mimiq_ops = {
            "i": "GateID()",
            "x": "GateX()",
            "y": "GateY()",
            "z": "GateZ()",
            "h": "GateH()",
            "t": "GateT()",
            "phase": "GateS()",
            "phaseinv": "GateSDG()",
            "cnot": "GateCX()",
            "cy": "GateCY()",
            "cz": "GateCZ()",
            "cx": "GateCX()",
            "swap": "GateSWAP()",
            "reset": "ResetZ()",
            "reset_0": "ResetZ()",
            "reset_1": ["ResetZ()", "GateX()"],
            "reset_+": "ResetX()",
            "reset_-": ["ResetX()", "GateZ()"],
            "reset_+i": "ResetY()",
            "reset_-i": ["ResetY()", "GateZ()"],
            "measurement": "MeasureZ()",
            "measure_z": "MeasureZ()",
            "measure_x": "MeasureX()",
            "measure_y": "MeasureY()",
        }

        return {
            op.name: partial(op_type_handlers[op.op_type], op=eka_to_mimiq_ops[op.name])
            for op in USUAL_QUANTUM_GATES | NONCLIFFORD_GATES_SIGNATURE
        } | {
            op.name: partial(op_type_handlers[op.op_type], op=op.name)
            for op in self.MIMIQ_CLASSICALLY_CONTROLLED_OPS | UTILS_SIGNATURE
        }

    def convert_circuit(
        self,
        input_circuit: Circuit,
    ) -> tuple[str, dict[Channel, int], dict[Channel, int]]:
        """Convert a Circuit to a MimiqCircuitAndRegisterMap.

        Parameters
        ----------
        input_circuit : Circuit
            The input circuit to convert.

        Returns
        -------
        MimiqCircuitAndRegisterMap
            The converted Mimiq circuit program and register map.
        """

        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit")

        if (
            not input_circuit.circuit and input_circuit.name not in self.operations_map
        ) or not input_circuit.channels:
            return "# empty input", {}, {}

        mimiq_program_lines = []

        mimiq_program_lines.append("# Mimiq Program Generated from Eka")

        init_instr, q_register, c_register = self.emit_init_instructions(input_circuit)

        mimiq_program_lines.extend(init_instr.splitlines())

        mimiq_program_lines.append("")

        unrolled_circuit = Circuit.unroll(input_circuit)

        for layer in unrolled_circuit:
            for eka_op in layer:
                if eka_op.name == "if-else_circuit":
                    eka_op = self.parse_if_operation(eka_op)
                if eka_op.name not in self.operations_map:
                    raise ValueError(f"Unsupported operation: {eka_op.name}")
                mimiq_program_lines.append(
                    self.emit_leaf_circuit_instruction(eka_op, q_register, c_register)
                )

        return "\n".join(mimiq_program_lines), q_register, c_register

    @staticmethod
    def parse_target_run_outcome(
        run_output: MimiqResult,
    ) -> dict[str, int | list[int]]:
        """Parse the run output of the target language into a dictionary mapping the
        eka channel labels to boolean values measured at each shot."""
        mimiq_result, channel_to_idx = run_output
        result = mimiq_result.histogram()
        bitstrings = [key for key, count in result.items() for _ in range(count)]

        result = {chan.label: [] for chan in channel_to_idx.keys()}

        if bitstrings == []:
            raise ValueError("No bitstrings found in the run output.")

        for channel in channel_to_idx.keys():
            idx = channel_to_idx[channel]
            result[channel.label] = [int(b[idx]) for b in bitstrings]

        return result

    def emit_init_instructions(
        self, input_circuit: Circuit
    ) -> tuple[str, dict[Channel, int], dict[Channel, int]]:
        """Provide the python code (as a string) to initializes the
        quantum and classical registers, and return the mapping from eka channel to
        register."""
        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        q_channels = sorted(
            [qc for qc in input_circuit.channels if qc.is_quantum()],
            key=lambda c: c.label,
        )
        c_channels = sorted(
            [cc for cc in input_circuit.channels if cc.is_classical()],
            key=lambda c: c.label,
        )
        q_register = {qc: idx for idx, qc in enumerate(q_channels)}
        c_register = {cc: idx for idx, cc in enumerate(c_channels)}
        # Initialize the Mimiq circuit and register map
        return (
            f"{self.circuit_varname} = {self.mimiq_import_prefix}Circuit()",
            q_register,
            c_register,
        )

    def emit_leaf_circuit_instruction(
        self,
        input_circuit: Circuit,
        quantum_channel_map: dict[Channel, int],
        classical_channel_map: dict[Channel, int],
    ) -> str:
        """Provide the python code (as a string) to emit an Eka instruction in the
        target language."""

        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if input_circuit.circuit and len(input_circuit.circuit) != 0:
            raise ValueError("The circuit must be an leaf circuit with channels.")

        if input_circuit.name not in self.operations_map:
            raise ValueError(f"Unsupported operation: {input_circuit.name}")

        q_targets = [
            str(quantum_channel_map[q_chan])
            for q_chan in input_circuit.channels
            if q_chan.is_quantum()
        ]

        c_targets = [
            classical_channel_map[c_chan]
            for c_chan in input_circuit.channels
            if c_chan.is_classical()
        ]

        if not input_circuit.name.startswith("classical_controlled_"):
            self._validate_ops_args(input_circuit.name, len(q_targets), len(c_targets))

        return self.operations_map[input_circuit.name](
            q_targets, c_targets, desc=input_circuit.description
        )

    def parse_if_operation(self, if_circuit: IfElseCircuit) -> Circuit:
        """
        Parse control flow operation, allowing only specific cases supported by MIMIQ.
        MIMIQ only supports conditional unitary operations.

        Parameters
        ----------
        if_circuit : IfElseCircuit
            The IfElseCircuit to parse.

        Returns
        -------
        Circuit
            A Circuit (gate) representing the parsed if operation in MIMIQ.
        """

        if not if_circuit.is_single_gate_conditioned:
            raise ValueError(
                "Unsupported operation for Mimiq conversion: "
                "Mimiq only supports single gate conditioned if-else circuits."
            )

        applied_unitary = if_circuit.if_circuit.circuit[0][0].name
        if applied_unitary in self.op_by_eka_name:
            if self.op_by_eka_name[applied_unitary] not in CLIFFORD_GATES_SIGNATURE:
                raise ValueError(
                    "Unsupported operation for Mimiq conversion: "
                    "Mimiq only supports single qubit unitary operations in if-else "
                    "circuits."
                )

        desc = ""
        match if_circuit.condition_circuit.name:
            case BoolOp.MATCH:
                desc = "1"
            case BoolOp.NOT:
                desc = "0"
            case BoolOp.AND:
                desc = "1" * len(if_circuit.condition_circuit.channels)
            case BoolOp.NOR:
                desc = "0" * len(if_circuit.condition_circuit.channels)
            case _:
                raise ValueError(
                    f"Unsupported bool operator for Mimiq"
                    f"conditional statement: {if_circuit.condition_circuit.name}"
                )
        return Circuit(
            name=f"classical_controlled_{applied_unitary}",
            channels=if_circuit.if_circuit.channels
            + if_circuit.condition_circuit.channels,
            description=desc,
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-lines
from loom.eka import Circuit, ChannelType, SyndromeCircuit, Stabilizer, PauliOperator
from loom.eka.utilities import Direction, Orientation, DiagonalDirection
from loom.eka.operations import MeasureBlockSyndromes
from loom.interpreter.applicator import measureblocksyndromes
from loom.interpreter import InterpretationStep

from loom_rotated_surface_code.code_factory import RotatedSurfaceCode
from .utilities import (
    generate_syndrome_extraction_circuits,
    find_relative_diagonal_direction,
)


def move_corners(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    corner_args: tuple[tuple[tuple[int, int, int], Direction, int], ...],
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Move the selected topological corners of the block in the specified direction. We
    assume that the initial state is a RotatedSurfaceCode with a single logical qubit
    and the topological corners are located at the geometric corners.

    This function can move he topological corners mutlitple times in a row, as long as
    the qubits selected are topological corners of the block.

    NOTE: This function automatically measures the syndromes of the final block once.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step before moving the corners
    block : RotatedSurfaceCode
        The initial block
    corner_args : tuple[tuple[tuple[int, int, int], Direction, int], ...]
        List of tuples containing the corner qubit to move, the direction in which to
        move it and the distance by which to move it
    same_timeslice : bool
        Flag to apply move_corners in the same timeslice as the previous operation(s).
    debug_mode: bool
        Flag to apply validation of the new block or not.

    Returns
    -------
    InterpretationStep
        The interpretation step after moving the corners
    """
    initial_block_name = block.unique_label

    # Collect the stabilizers that haven't been measured yet but are required to
    # update the logical operators
    stabilizers_required_for_future_update = {"X": (), "Z": ()}

    # Iterate over the corners to move
    for corner_qubit, move_direction, how_far in corner_args:
        # Store the previous logical_operators to check if we need to inherit updates
        previous_logical_operators = (
            block.logical_x_operators[0],
            block.logical_z_operators[0],
        )

        # Rotate corners
        interpretation_step, past_stabilizers, future_stabilizers = move_corner(
            interpretation_step,
            block,
            corner_qubit,
            move_direction,
            how_far,
        )
        stabilizers_required_for_future_update["X"] += future_stabilizers[0]
        stabilizers_required_for_future_update["Z"] += future_stabilizers[1]
        # Change the reference to the block
        block = interpretation_step.get_block(initial_block_name)

        # Add updates from the past stabilizers
        for pauli_type, past_stabs, logical_operator in zip(
            ("X", "Z"),
            past_stabilizers,
            (block.logical_x_operators[0], block.logical_z_operators[0]),
            strict=True,
        ):
            cbits = tuple(
                cbit
                for stab in past_stabs
                for cbit in interpretation_step.get_prev_syndrome(
                    stab.uuid, block.uuid
                )[0].measurements
            )
            # We only want to inherit updates once every time the logical operator is
            # changed
            inherit_updates = not logical_operator in previous_logical_operators
            interpretation_step.update_logical_operator_updates_MUT(
                pauli_type,
                logical_operator.uuid,
                cbits,
                inherit_updates=inherit_updates,
            )

    # Measure the syndromes of the new block
    interpretation_step = measureblocksyndromes(
        interpretation_step=interpretation_step,
        operation=MeasureBlockSyndromes(initial_block_name),
        same_timeslice=same_timeslice,
        debug_mode=debug_mode,
    )
    # Add the logical updates from the remaining stabilizers
    # Find the previous Cbit associated with the stabilizers required for the updates
    for pauli_type, future_stabs in stabilizers_required_for_future_update.items():
        logical_operator = (
            interpretation_step.get_block(initial_block_name).logical_x_operators[0]
            if pauli_type == "X"
            else interpretation_step.get_block(initial_block_name).logical_z_operators[
                0
            ]
        )
        cbits = tuple(
            cbit
            for stab in future_stabs
            for cbit in interpretation_step.get_prev_syndrome(stab.uuid, block.uuid)[
                0
            ].measurements
        )
        interpretation_step.update_logical_operator_updates_MUT(
            pauli_type, logical_operator.uuid, cbits, inherit_updates=False
        )

    return interpretation_step


def move_corner(  # pylint: disable=too-many-locals, too-many-statements
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
    move_direction: Direction,
    how_far: int,
) -> tuple[
    InterpretationStep,
    tuple[tuple[Stabilizer, ...], tuple[Stabilizer, ...]],
    tuple[tuple[Stabilizer, ...], tuple[Stabilizer, ...]],
]:
    """
    Move the selected topological corner of the block in the specified direction. We
    assume that the initial state is a RotatedSurfaceCode with a single logical qubit
    and the topological corners are located at the geometric corners.

    The algorithm is the following:

    - A.) STABILIZERS

        - A.1) Cut the corner stabilizer if a 2-body stabilizer is involved
        - A.2) Generate the two-body stabilizers to be added
        - A.3) Find which stabilizers are untouched on the modified boundary
        - A.4) Compute the evolution of the stabilizers
        - A.5) Construct the set of kept bulk stabilizers
        - A.6) Construct the new set of stabilizers

    - B.) LOGICAL OPERATORS

        - B.1) Move the logical operators so they stay on the right boundary or
            extend/contract them
        - B.2) Update the logical operators
        - B.3) Collect the necessary stabilizers for logical operator updates

    - C.) SYNDROME CIRCUITS

        - C.1) Find corner configuration of new block
        - C.2) Generate new syndrome circuits and stabilizer schedules based on corner
            config
        - C.3) Update stabilizer evolution

    - D.) CIRCUIT

        - D.1) Measure the corner if needed
        - D.2) Create stabilizer and logical operator updates

    - E.) BUILD BLOCK

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step before moving the corner
    block : RotatedSurfaceCode
        The initial block
    corner_qubit : tuple[int, int, int],
        Corner to move
    move_direction : Direction
        Direction in which to move the corner
    how_far : int
        Distance by which to move the corner

    Returns
    -------
    tuple[
        InterpretationStep,
        tuple[tuple[Stabilizer, ...], tuple[Stabilizer, ...]],
        tuple[tuple[Stabilizer, ...], tuple[Stabilizer, ...]],
    ]
        The interpretation step after moving the corner, the set of X and Z stabilizers
         removed during the move that are required for logical updates and the set of X
         and Z stabilizers added during the move that are required for logical updates.
    """
    # We first compute some important values:
    # The boundary associated with the corner
    corner_directions = get_associated_boundaries(block, corner_qubit)
    # The boundary modified depends on the corner and the direction of the movement
    modified_boundary_direction = find_new_boundary_direction(
        corner_directions, move_direction
    )
    # The boundary that is extended, e.g. if the top left corner is moved to the right,
    # the left boundary is extended
    unit_vector = move_direction.to_vector()

    two_body_is_included = is_2_body_included(
        block, corner_qubit, modified_boundary_direction
    )
    # A) STABILIZERS
    #   A.1) Cut the corner stabilizer if a 2-body stabilizer is involved
    #   A.2) Generate the two-body stabilizers to be added
    #   A.3) Find which stabilizers are untouched on the modified boundary
    #   A.4) Compute the evolution of the stabilizers
    (
        kept_boundary_stabilizers,
        new_boundary_stabilizers,
        cut_stabilizer,
        stabilizer_evolution,
    ) = find_new_boundary_stabilizers(
        block,
        corner_qubit,
        modified_boundary_direction,
        unit_vector,
        how_far,
        two_body_is_included,
    )
    old_stabs_to_be_removed = [
        stab
        for stab in block.boundary_stabilizers(modified_boundary_direction)
        if stab not in kept_boundary_stabilizers
    ]
    interpretation_step.stabilizer_evolution.update(stabilizer_evolution)
    #   A.5) Construct the set of kept bulk stabilizers
    kept_bulk_stabilizers = [
        stab
        for stab in block.bulk_stabilizers
        if not (corner_qubit in stab.data_qubits and two_body_is_included)
    ]
    #   A.6) Construct the new set of stabilizers
    new_stabilizers = (
        kept_boundary_stabilizers + new_boundary_stabilizers + kept_bulk_stabilizers
    )
    if cut_stabilizer is not None:
        new_stabilizers.append(cut_stabilizer)

    # B) LOGICAL OPERATORS
    #   B.1) Move the logical operators so they stay on the right boundary or
    #        extend/contract them
    new_x_op, new_z_op, log_x_evolution_dict, log_z_evolution_dict = (
        move_corner_logical_operators(
            block=block,
            corner_qubit=corner_qubit,
            unit_vector=unit_vector,
            how_far=how_far,
            new_stabs_to_be_added=new_boundary_stabilizers,
            old_stabs_to_be_removed=old_stabs_to_be_removed,
        )
    )
    #   B.2) Update the logical operator evolution
    interpretation_step.logical_x_evolution.update(log_x_evolution_dict)
    interpretation_step.logical_z_evolution.update(log_z_evolution_dict)

    #   B.3) Collect the necessary stabilizers for logical operator updates
    # Collect the removed stabilizers that are required for logical updates
    # (their syndromes should have been measured before)
    past_x_stabilizers = tuple(
        stab
        for stab in old_stabs_to_be_removed
        for pauli_op_id in interpretation_step.logical_x_evolution.get(
            new_x_op.uuid, ()
        )
        if stab.uuid == pauli_op_id
    )
    past_z_stabilizers = tuple(
        stab
        for stab in old_stabs_to_be_removed
        for pauli_op_id in interpretation_step.logical_z_evolution.get(
            new_z_op.uuid, ()
        )
        if stab.uuid == pauli_op_id
    )
    past_stabilizers = (past_x_stabilizers, past_z_stabilizers)

    # Collect the new stabilizers that are required for logical updates
    # (their syndromes have not been measured yet)
    future_x_stabilizers = tuple(
        stab
        for stab in new_boundary_stabilizers
        for evolution in interpretation_step.logical_x_evolution.values()
        for pauli_op_id in evolution
        if stab.uuid == pauli_op_id
    )
    future_z_stabilizers = tuple(
        stab
        for stab in new_boundary_stabilizers
        for evolution in interpretation_step.logical_z_evolution.values()
        for pauli_op_id in evolution
        if stab.uuid == pauli_op_id
    )
    future_stabilizers = (future_x_stabilizers, future_z_stabilizers)

    # C) SYNDROME CIRCUITS
    # C.1) Create the mock block that describes the final geometry
    mock_block = RotatedSurfaceCode(
        unique_label=block.unique_label,
        stabilizers=new_stabilizers,
        logical_x_operators=[new_x_op],
        logical_z_operators=[new_z_op],
    )

    # C.2) Select the starting qubit diagonal direction
    # By default use TOP-RIGHT direction.
    # For type 2 and 3, the starting qubit should be away from the new boundary and
    # towards the move boundary in the next move corner to ensure fault-tolerance
    config, pivot_corners = mock_block.config_and_pivot_corners
    starting_diag_direction = DiagonalDirection.TOP_RIGHT
    if config in (2, 3, 4):
        # Type 2 and 3 (U and L)
        # Extract geometry information
        is_move_direction_horizontal = (
            move_direction.to_orientation() == Orientation.HORIZONTAL
        )
        is_horizontal = mock_block.is_horizontal
        is_move_direction_long_edge = is_move_direction_horizontal == is_horizontal
        long_end_corner = pivot_corners[0]
        short_end_corner = pivot_corners[3]

        # Comparison of these two corners will determine the starting diagonal direction
        top_left_corner = mock_block.upper_left_qubit
        compare_corner = (
            long_end_corner if is_move_direction_long_edge else short_end_corner
        )
        starting_diag_direction = find_relative_diagonal_direction(
            top_left_corner, compare_corner
        )

    #   C.2) Find syndrome circuits using the mock block
    new_syndrome_circuits, new_stab_to_circuit = generate_syndrome_extraction_circuits(
        mock_block, starting_diag_direction
    )

    #   C.3) Get the cut corner syndrome circuit (not calculated by the function above)
    if isinstance(cut_stabilizer, Stabilizer):
        cut_syndrome_circuit = cut_corner_syndrome_circuit(
            block, cut_stabilizer, corner_qubit, corner_directions
        )
        new_syndrome_circuits += (cut_syndrome_circuit,)
        new_stab_to_circuit.update({cut_stabilizer.uuid: cut_syndrome_circuit.uuid})

    # D) CIRCUIT
    #   D.1) Measure the corner if needed
    circuit_output = move_corner_circuit(
        interpretation_step=interpretation_step,
        block=block,
        corner_qubit=corner_qubit,
        cut_stabilizer=cut_stabilizer,
        move_direction=move_direction,
        how_far=how_far,
    )
    if circuit_output is not None:
        corner_circuit, new_cbit = circuit_output
        interpretation_step.append_circuit_MUT(corner_circuit)
        #   D.2) Create stabilizer and logical operator updates
        interpretation_step.stabilizer_updates[cut_stabilizer.uuid] = (new_cbit,)
        if corner_qubit in block.logical_x_operators[0].data_qubits:
            interpretation_step.logical_x_operator_updates[new_x_op.uuid] = (new_cbit,)
        if corner_qubit in block.logical_z_operators[0].data_qubits:
            interpretation_step.logical_z_operator_updates[new_z_op.uuid] = (new_cbit,)

    syndrome_circuits_in_used = tuple(new_stab_to_circuit.values())
    new_syndrome_circuits = tuple(
        syndrome_circuit
        for syndrome_circuit in new_syndrome_circuits
        if syndrome_circuit.uuid in syndrome_circuits_in_used
    )

    # E) BUILD BLOCK
    new_block = RotatedSurfaceCode(
        unique_label=block.unique_label,
        stabilizers=new_stabilizers,
        logical_x_operators=[new_x_op],
        logical_z_operators=[new_z_op],
        syndrome_circuits=new_syndrome_circuits,
        stabilizer_to_circuit=new_stab_to_circuit,
    )
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,), old_blocks=(block,)
    )

    return interpretation_step, past_stabilizers, future_stabilizers


def get_associated_boundaries(
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
) -> tuple[Direction, ...]:
    """Get the boundary direction(s) associated with the corner qubit

    Parameters
    ----------
    block : RotatedSurfaceCode
        Block of rotated surface code
    corner_qubit : tuple[int, int, int]
        Corner qubit

    Returns
    -------
    tuple[Direction, ...]
        Boundary direction(s) associated with the corner qubit

    Raises
    ------
    ValueError
        If the corner given is not a topological corner of the rotated surface code
        block, i.e. the product of stabilizers at this point is not Y.
    """

    # Check if the corner is a topological corner
    if corner_qubit not in block.topological_corners:
        raise ValueError(
            f"The selected corner qubit {corner_qubit} is not a topological corner of"
            + f" the block `{block.unique_label}`"
        )

    # Find the boundaries associated with the corner
    corner_directions = tuple(
        direction
        for direction in Direction
        if corner_qubit in block.boundary_qubits(direction)
    )
    # Should not happen unless there are defects in the block, our assumptions break
    # down here
    if len(corner_directions) == 0:
        raise ValueError(
            f"The corner qubit {corner_qubit} is not associated with any boundary"
        )

    return corner_directions


def find_new_boundary_direction(
    corner_boundaries: tuple[Direction, ...],
    move_direction: Direction,
) -> Direction:
    """Find the modified boundary direction that corresponds to moving the corner in the
    specified direction. E.g. moving the top left corner towards the right modifies the
    top boundary.

    Parameters
    ----------
    corner_boundaries : tuple[Direction, ...]
        Description of the corner in terms of the geometric directions. E.g. (TOP, LEFT)
        describes the top left corner. A topological corner can be included in a single
        boundary or two boundaries.
    move_direction : Direction
        Direction in which to move the corner

    Returns
    -------
    Direction
        Direction of the boundary that will be modified
    """
    if (n_boundaries := len(corner_boundaries)) not in (1, 2):
        raise ValueError(
            f"Invalid number of corner boundaries: {n_boundaries}, "
            "must be either 1 or 2"
        )
    if set(corner_boundaries) in (
        {Direction.TOP, Direction.BOTTOM},
        {Direction.LEFT, Direction.RIGHT},
    ):
        raise ValueError(
            f"Invalid corner boundaries: {corner_boundaries}, they must be "
            + "orthogonal, e.g. (TOP, LEFT) or a single direction."
        )
    if move_direction in corner_boundaries:
        raise ValueError(
            f"Cannot move the corner {corner_boundaries} towards the "
            f"{move_direction.value}"
        )

    return next(
        direction
        for direction in corner_boundaries
        if direction != move_direction.opposite()
    )


def is_2_body_included(
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
    boundary: Direction,
) -> bool:
    """Check if the corner qubit is included in a 2-body stabilizer that is part of the
    chosen boundary.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block
    corner_qubit : tuple[int, int, int]
        Corner qubit to check
    boundary : Direction
        Boundary to check

    Returns
    -------
    bool
        True if the corner qubit is included in a 2-body stabilizer that is part of the
        chosen boundary. False otherwise.
    """
    if corner_qubit not in (boundary_qubits := block.boundary_qubits(boundary)):
        raise ValueError(
            f"The corner qubit {corner_qubit} is not part of the {boundary.value} "
            f"boundary"
        )
    return any(
        stab
        for stab in block.stabilizers
        if corner_qubit in stab.data_qubits
        and all(q in boundary_qubits for q in stab.data_qubits)
    )


def cut_corner_stabilizer(
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
) -> Stabilizer:
    """Create a new 3-body stabilizer by cutting the corner qubit from initial 4-body
    stabilizer.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block
    corner_qubit : tuple[int, int, int]
        Corner qubit that is cut

    Returns
    -------
    Stabilizer
        New stabilizer with the corner qubit cut
    """
    # We assume that the corner is indeed a corner
    stab_to_cut = next(
        stab
        for stab in block.stabilizers
        if corner_qubit in stab.data_qubits and len(stab.data_qubits) == 4
    )
    # Keep the order of the data qubits consistent with the original stabilizer
    new_pauli, new_qubits = zip(
        *(
            (pauli, qubit)
            for pauli, qubit in zip(
                stab_to_cut.pauli, stab_to_cut.data_qubits, strict=True
            )
            if qubit != corner_qubit
        ),
        strict=True,
    )
    corner_stabilizer = Stabilizer(
        pauli="".join(new_pauli),
        data_qubits=new_qubits,
        ancilla_qubits=stab_to_cut.ancilla_qubits,
    )
    return corner_stabilizer


def cut_corner_syndrome_circuit(
    block: RotatedSurfaceCode,
    cut_stabilizer: Stabilizer,
    corner_qubit: tuple[int, int, int],
    which_corner: tuple[Direction, Direction],
) -> SyndromeCircuit:
    """Cut the corner qubit from the syndrome circuit that measures the stabilizer to
    cut. Note that the name of the SyndromeCircuit is always
    vertical_dir-horizontal_dir-pauli.

    Parameters
    ----------
    block: RotatedSurfaceCode
        Initial block
    cut_stabilizer: Stabilizer
        Stabilizer with the corner qubit cut out
    corner_qubit : tuple[int, int, int]
        Corner qubit to cut
    which_corner : tuple[Direction, Direction]
        Position of the corner qubit to cut

    Returns
    -------
    SyndromeCircuit
        Syndrome circuit with the corner qubit cut
    """
    stab_to_cut = next(
        stab for stab in block.bulk_stabilizers if corner_qubit in stab.data_qubits
    )

    # Order the directions
    vertical_direction = next(
        dir for dir in which_corner if dir in {Direction.TOP, Direction.BOTTOM}
    )
    horizontal_direction = next(
        dir for dir in which_corner if dir in {Direction.LEFT, Direction.RIGHT}
    )

    # Create a syndrome circuit with the corner qubit cut
    cut_syndrome_circuit = RotatedSurfaceCode.generate_syndrome_circuit(
        pauli=cut_stabilizer.pauli,
        padding=[stab_to_cut.data_qubits.index(corner_qubit)],
        name=f"{vertical_direction}-{horizontal_direction}-{cut_stabilizer.pauli}",
    )

    return cut_syndrome_circuit


def generate_updated_2_body_stabilizers(
    old_corner_qubit: tuple[int, int, int],
    new_boundary_direction: Direction,
    unit_vector: tuple[int, int],
    how_far: int,
    pauli_type: str,
) -> list[Stabilizer]:
    """Generate the new boundary stabilizers that are created when moving the corner.

    Parameters
    ----------
    old_corner_qubit : tuple[int, int, int]
        Old corner qubit
    new_boundary_direction : Direction
        Direction of the modified boundary
    unit_vector : tuple[int, int]
        Unit vector of the movement
    how_far : int
        Distance by which the corner is moved
    pauli_type : str
        Pauli type of the new stabilizers

    Returns
    -------
    list[Stabilizer]
        List of new stabilizers
    """
    if (abs(unit_vector[0]), abs(unit_vector[1])) not in ((1, 0), (0, 1)):
        raise ValueError("Invalid unit vector")
    orientation = Orientation.from_vector(unit_vector)

    is_bottom_or_right = new_boundary_direction in (Direction.RIGHT, Direction.BOTTOM)

    adjusted_old_corner_qubit = (
        old_corner_qubit[0] + unit_vector[0] * how_far % 2,  # If odd, the corner is cut
        old_corner_qubit[1] + unit_vector[1] * how_far % 2,  # If odd, the corner is cut
        0,
    )
    final_stab_position = (
        old_corner_qubit[0] + unit_vector[0] * (how_far - 1),
        old_corner_qubit[1] + unit_vector[1] * (how_far - 1),
        0,
    )
    # We create the stabilizer from top to bottom or left to right
    initial_position = min(
        (adjusted_old_corner_qubit, final_stab_position), key=lambda x: x[0] + x[1]
    )

    new_stabs = RotatedSurfaceCode.generate_weight2_stabs(
        pauli=pauli_type * 2,
        initial_position=initial_position,
        num_stabs=how_far // 2,
        orientation=orientation,
        is_bottom_or_right=is_bottom_or_right,
    )
    return new_stabs


def find_new_boundary_stabilizers(
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
    modified_boundary_direction: Direction,
    unit_vector: tuple[int, int],
    how_far: int,
    two_body_is_included: bool,
) -> tuple[list[Stabilizer], list[Stabilizer], Stabilizer | None, dict[str, str]]:
    """Finds the stabilizers of the new modified boundary after moving the corner. Some
    stabilizers of the old boundary are not modified and returned as well. If a 2-body
    stabilizer is involved in the selected corner and the modified boundary, the corner
    stabilizer is cut and the stabilizer that is moved is returned.
    It returns, the kept boundaries stabilizers, the newly created ones, the cut bulk
    stabilizer and the stabilizer evolution.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block
    corner_qubit : tuple[int, int, int]
        Corner qubit to move
    modified_boundary_direction : Direction
        Direction of the modified boundary
    unit_vector : tuple[int, int]
        Unit vector describing the movement of the corner
    how_far : int
        Distance by which to move the corner
    two_body_is_included : bool
        Is there a two-body stabilizer involved in the corner and the modified boundary
        in the direction of movement.

    Returns
    -------
    tuple[list[Stabilizer], list[Stabilizer], Stabilizer | None, dict[str, str]]
        Kept boundaries stabilizers, new boundary stabilizers, cut stabilizer,
        stabilizer evolution.

    Raises
    ------
    ValueError
        If the distance is not odd when a 2-body stabilizer is involved or if the
        distance is not even when there is no 2-body stabilizer involved.
    """
    qubits_to_update = [
        (
            corner_qubit[0] + i * unit_vector[0],
            corner_qubit[1] + i * unit_vector[1],
            0,
        )
        for i in range(how_far + 1)  # Old corner and new corners are included
    ]
    kept_boundary_stabilizers = [
        stab
        for stab in block.all_boundary_stabilizers
        if not all(q in qubits_to_update for q in stab.data_qubits)
    ]
    # Check if a 2-body stabilizer is involved in the selected corner and the modified
    # boundary
    # If it is, we need to cut the geometric corner to move the topological corner
    if two_body_is_included:
        # Only odd distances are allowed
        if how_far % 2 == 0:
            raise ValueError(
                "Only odd distances are allowed for moving the corner in the direction"
                " of the 2-body stabilizer"
            )
        stab_to_cut = next(
            stab for stab in block.bulk_stabilizers if corner_qubit in stab.data_qubits
        )
        # Cut the corner
        cut_stabilizer = cut_corner_stabilizer(block, corner_qubit)

        stabilizer_evolution = {cut_stabilizer.uuid: (stab_to_cut.uuid,)}

    else:
        # Only even distances are allowed
        if how_far % 2 == 1:
            raise ValueError(
                "Only even distances are allowed for moving the corner in the direction"
                " where there is no 2-body stabilizer"
            )

        cut_stabilizer = None
        stabilizer_evolution = {}
        # Just change the 2-body stabilizers

    try:
        new_boundary_stabilizers = generate_updated_2_body_stabilizers(
            corner_qubit,
            modified_boundary_direction,
            unit_vector,
            how_far,
            block.boundary_type(modified_boundary_direction),
        )
    except RuntimeError:
        # If there is more than 1 boundary type
        # Do this to get the exact boundary type.
        boundary_pauli_charges = list(
            set(block.pauli_charges[each_qubit] for each_qubit in qubits_to_update)
            - set("Y")
        )
        new_boundary_stabilizers = generate_updated_2_body_stabilizers(
            corner_qubit,
            modified_boundary_direction,
            unit_vector,
            how_far,
            boundary_pauli_charges[0],
        )

    return (
        kept_boundary_stabilizers,
        new_boundary_stabilizers,
        cut_stabilizer,
        stabilizer_evolution,
    )


def move_corner_logical_operators(  # pylint: disable=too-many-locals
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
    unit_vector: tuple[int, int],
    how_far: int,
    old_stabs_to_be_removed: list[Stabilizer],
    new_stabs_to_be_added: list[Stabilizer],
) -> tuple[PauliOperator, PauliOperator, dict[str, str], dict[str, str]]:
    """Modify the logical operators of the block after moving the corner. There is no
    guarantee that the distance of the block is preserved after moving the corners,
    therefore no guarantee on the length of the operators.
    Note that the logical operator is not modified if the movement does not affect the
    commutation properties, it may not be of minimal weight anymore. The distance of
    the code may not be preserved.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block
    corner_qubit : tuple[int, int, int]
        Corner qubit to move
    unit_vector : tuple[int, int]
        Unit vector describing the movement of the corner
    how_far : int
        Distance by which to move the corner
    old_stabs_to_be_removed : list[Stabilizer]
        Stabilizers that are removed by the corner movement.
    new_stabs_to_be_added : list[Stabilizer]
        Stabilizers that are added by the corner movement.

    Returns
    -------
    tuple[PauliOperator, PauliOperator, dict[str, str], dict[str, str]]
        New logical X operator, new logical Z operator, logical X evolution dictionary,
        logical Z evolution dictionary.
    """
    qubits_to_update = [
        (
            corner_qubit[0] + i * unit_vector[0],
            corner_qubit[1] + i * unit_vector[1],
            0,
        )
        for i in range(how_far + 1)  # Both new and old corners are included
    ]

    # Represent a product of pauli P1*P2 = pauli_product[P1][P2] (up to phase)
    pauli_product = {
        "X": {"X": "I", "Z": "Y", "Y": "Z", "I": "X"},
        "Z": {"X": "Y", "Z": "I", "Y": "X", "I": "Z"},
        "Y": {"X": "Z", "Z": "X", "Y": "I", "I": "Y"},
        "I": {"X": "X", "Z": "Z", "Y": "Y", "I": "I"},
    }

    # Make sure we are working with a single logical qubit
    if len(block.logical_x_operators) > 1 or len(block.logical_z_operators) > 1:
        raise ValueError("Only a single logical qubit is supported")
    # NOTE we do not support Y paulis

    # Update the logical operators
    old_x_op = block.logical_x_operators[0]
    old_z_op = block.logical_z_operators[0]

    # Case 1 - The logical operator is not affected if the qubits are not included in
    # the movement.
    # Case 2 : All qubits to be modified are included in the operator: we shorten
    # the operator but make sure the new corner (qubits_to_update[-1]) is included.
    # Case 3 - The operator is on the boundary and needs to be modified. We use the
    # anchor (see inline documentation below)on the boundary to determine where to
    # modify the operator and drag the operator to the new topological corner.
    # Refer to the tests for examples of each case.

    new_ops, log_evolution_dicts = [], []
    for old_op in (old_x_op, old_z_op):
        logical_op_dict = {
            q: p for p, q in zip(old_op.pauli, old_op.data_qubits, strict=True)
        }
        qubits_included_op = set(qubits_to_update).intersection(set(old_op.data_qubits))
        # Case 1: The logical operator is not affected
        if len(qubits_included_op) == 0:
            new_op = old_op
            log_evolution_dict = {}
        # Case 2 : All qubits to be modified are included in the operator: we shorten
        # the operator but make sure the new corner (qubits_to_update[-1]) is included
        elif set(qubits_to_update) == qubits_included_op:
            new_op = PauliOperator(
                pauli="".join(
                    p
                    for (q, p) in logical_op_dict.items()
                    if q not in qubits_to_update[:-1]
                ),
                data_qubits=tuple(
                    q
                    for (q, p) in logical_op_dict.items()
                    if q not in qubits_to_update[:-1]
                ),
            )
            log_evolution_dict = {
                new_op.uuid: (old_op.uuid,)
                + tuple(stab.uuid for stab in new_stabs_to_be_added)
            }
        # Case 3: The operator is on the boundary and needs to be modified, we find
        # the anchor and "drag" the operator to the new corner
        else:
            # Find the anchor on the boundary
            if corner_qubit in old_op.data_qubits:
                # The anchor is either the old corner
                anchor = corner_qubit
            else:
                # Or the first qubit that breaks the commutation relation with the new
                # stabilizers, please refer to tests for an example
                old_op_data_qubits_set = set(old_op.data_qubits)
                intersection_with_stabs = [
                    set(stab.data_qubits).intersection(old_op_data_qubits_set)
                    for stab in new_stabs_to_be_added
                ]
                anchor = next(
                    next(iter(intersection))  # We reasonably assume a single anchor
                    for intersection in intersection_with_stabs
                    if len(intersection) == 1
                )
            old_stab_dict = {
                q: p
                for stab in old_stabs_to_be_removed
                for p, q in zip(stab.pauli, stab.data_qubits, strict=True)
            }
            # The operator is not modified until we touch its anchor
            start_modify_index = qubits_to_update.index(anchor)
            # The operator is only modified by the old stabilizers located after the
            # anchor
            included_old_stabs = [
                stab
                for stab in old_stabs_to_be_removed
                if all(
                    q in qubits_to_update[start_modify_index:] for q in stab.data_qubits
                )
            ]
            # Modify the operator for each qubit that is updated
            for qubit in qubits_to_update[start_modify_index:]:
                # Create a dictionary of the logical operator including new qubits
                if qubit not in logical_op_dict.keys():
                    logical_op_dict[qubit] = "I"
                # Multiply the operator by the old stabilizers that are removed
                if qubit in old_stab_dict.keys():
                    logical_op_dict[qubit] = pauli_product[logical_op_dict[qubit]][
                        old_stab_dict[qubit]
                    ]
            new_op = PauliOperator(
                pauli="".join(p for p in logical_op_dict.values() if p != "I"),
                data_qubits=tuple(q for (q, p) in logical_op_dict.items() if p != "I"),
            )
            log_evolution_dict = {
                new_op.uuid: (old_op.uuid,)
                + tuple(stab.uuid for stab in included_old_stabs)
            }
        new_ops.append(new_op)
        log_evolution_dicts.append(log_evolution_dict)

    new_x_op, new_z_op = new_ops  # pylint: disable=unbalanced-tuple-unpacking
    # pylint: disable=unbalanced-tuple-unpacking
    log_x_evolution_dict, log_z_evolution_dict = log_evolution_dicts

    return new_x_op, new_z_op, log_x_evolution_dict, log_z_evolution_dict


def move_corner_circuit(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    corner_qubit: tuple[int, int, int],
    cut_stabilizer: Stabilizer | None,
    move_direction: Direction,
    how_far: int,
):
    """Create the circuit that measures out the corner if needed. Returns the
    measurement circuit and the cbit that contains the measurement of the corner or
    None if the corner isn't cut.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step before moving the corner
    block : RotatedSurfaceCode
        The initial block
    corner_qubit : tuple[int, int, int],
        Corner to move
    cut_stabilizer : Stabilizer | None
        Stabilizer after being cut (only if a 2-body is involved in the movement, else
        None)
    move_direction : Direction
        Direction in which to move the corner
    how_far : int
        Distance by which to move the corner

    Returns
    -------
    tuple[Circuit, Cbit] | None
        Circuit measuring the corner qubit and Cbit that contains the corner measurement
        or None if the corner isn't cut.
    """
    if isinstance(cut_stabilizer, Stabilizer):
        stab_to_cut = next(
            stab for stab in block.bulk_stabilizers if corner_qubit in stab.data_qubits
        )
        pauli = next(
            p
            for p, q in zip(stab_to_cut.pauli, stab_to_cut.data_qubits, strict=True)
            if q == corner_qubit
        )
        measure_basis = "" if pauli == "Z" else f"_{pauli}"
        q_chan = interpretation_step.get_channel_MUT(str(corner_qubit))
        c_chan = interpretation_step.get_channel_MUT(
            f"c_{corner_qubit}", ChannelType.CLASSICAL
        )
        cbit = interpretation_step.get_new_cbit_MUT(f"c_{corner_qubit}")
        circuit = Circuit(
            name=f"moving corner {corner_qubit} to {move_direction} by {how_far}",
            circuit=((Circuit(f"measure{measure_basis}", channels=[q_chan, c_chan]),),),
        )

        return (circuit, cbit)

    return None
